package test;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import javax.transaction.NotSupportedException;
import javax.transaction.SystemException;
import javax.transaction.TransactionManager;

import org.apache.log4j.Logger;
import org.infinispan.Cache;
import org.infinispan.manager.DefaultCacheManager;
import org.infinispan.manager.EmbeddedCacheManager;
import org.testng.annotations.*;

public class TransactionLocksReleaseTest {

    private Logger log = Logger.getLogger(TransactionLocksReleaseTest.class);
    private EmbeddedCacheManager embeddedCacheManager;
    private TransactionManager tm;

    @BeforeMethod
    public void init() throws IOException {

        embeddedCacheManager = new DefaultCacheManager("local_config.xml", false);
        embeddedCacheManager.start();

        Cache cache = null;
        cache = embeddedCacheManager.getCache(SimpleEntity.class.getSimpleName());
        cache.start();
        cache = embeddedCacheManager.getCache(ComplexEntity.class.getSimpleName());
        cache.start();
        embeddedCacheManager.getCache().start();
        tm = embeddedCacheManager.getCache().getAdvancedCache().getTransactionManager();
    }

    @org.testng.annotations.Test
    public void testReleasingLocks() throws Exception {

        int threads = 2;

        ExecutorService executorService = Executors.newFixedThreadPool(threads);

        int groupsCount = 100;

        // batch of entities to update
        final List<ComplexEntity> groups = new ArrayList<ComplexEntity>();
        for(int i=0; i<groupsCount; i++){
            groups.add(new ComplexEntity(30+i, "group"+30+i, "group"));
        }

        // run threads that will update same batch data
        for (int i = 0; i < threads; i++) {
            final int j = i;
            executorService.submit(new Runnable() {

                @Override
                public void run() {
                    SimpleEntity entity = new SimpleEntity(1, "name", "surname " + j, false, 1000L, false);

                    try {
                        beginTransaction();
                        updateData(entity, groups);
                        commitTransaction();
                    } catch (Exception e) {
                        log.error("Thread failed", e);
                        rollbackTransaction();
                    }
                    log.info("------ AFTER EXCEPTION ----------");
                    beginTransaction();
                    updateData(entity, groups);
                    commitTransaction();
                }
            });
        }

        executorService.awaitTermination(30, TimeUnit.SECONDS);

        log.info("------ MAIN THREAD ----------");

        SimpleEntity entity = new SimpleEntity(1, "name", "surname", false, 1000L, false);

        // try to do data update several times with huge delays
        int tries = 5;
        for(int i=0; i<tries; i++) {
            log.debug("Update, try " + i);
            try {
                beginTransaction();
                updateData(entity, groups);
                commitTransaction();
            }catch(Exception e){
                log.error("Update " + i + " failed:\n", e);
            }
            Thread.sleep(15000);
        }

        Thread.sleep(60000);
        embeddedCacheManager.stop();
    }

    private void updateData(SimpleEntity user, List<ComplexEntity> groups){

        embeddedCacheManager.getCache(SimpleEntity.class.getSimpleName()).put(user.getPrimaryKey(), user);

        for(ComplexEntity group: groups ){
            embeddedCacheManager.getCache(ComplexEntity.class.getSimpleName()).remove(group.getPrimaryKey());
        }
        for(ComplexEntity group: groups ){
            embeddedCacheManager.getCache(ComplexEntity.class.getSimpleName()).put(group.getPrimaryKey(), group);
        }
    }

    private void beginTransaction() {
        log.debug("Starting new transaction.");
        try {
            tm.begin();
        } catch (NotSupportedException e) {
            String eMsg = "Unable to begin transaction: " + e.getMessage();
            log.error(eMsg, e);
            throw new RuntimeException(eMsg, e);
        } catch (SystemException e) {
            String eMsg = "Unable to begin transaction: " + e.getMessage();
            log.error(eMsg, e);
            throw new RuntimeException(eMsg, e);
        }
    }

    private void commitTransaction() {
        log.debug("Transaction commit.");
        try {
            tm.commit();
        } catch (Exception e) {
            String eMsg = "Unable to commit transaction: " + e.getMessage();
            log.error(eMsg, e);
            throw new RuntimeException(eMsg, e);
        }
    }

    private boolean rollbackTransaction() {
        log.debug("Transaction rollback.");
        try {
            tm.rollback();
        } catch (Exception e) {
            String eMsg = "Unable to rollback transaction: " + e.getMessage();
            log.error(eMsg, e);
            return false;
        }
        return true;
    }
}
